/* dict.c */

# include <stdio.h>

# define DCTSIZ 8192
/* # define DCTSIZ 1024 */
# define ALPSIZ 128
# define BITSWD 32
# define NUMNOD (DCTSIZ - ALPSIZ)
# define VECSIZ ((NUMNOD + BITSWD - 1) / BITSWD)
# define PERIOD 500
# define BUMP 1009

extern char *malloc();

char *dict[DCTSIZ];		/* string dictionary */
int count;			/* dict entries in use */
int prev;			/* # of last output entry */

long hashed[NUMNOD];		/* **** EXPERIMENTAL **** */
int leng[NUMNOD];		/* **** EXPERIMENTAL **** */
int ucount[NUMNOD];		/* **** EXPERIMENTAL **** */
int left[NUMNOD];		/* **** EXPERIMENTAL **** */
int right[NUMNOD];		/* **** EXPERIMENTAL **** */

static int rover;		/* NUR reap scan ptr */
static int clock;		/* NUR clear timer */
static int vector[VECSIZ];	/* NUR bitvector */

setbit(n)			/* set a bit in NUR bitvector (global) */
register int n;
{
	if( n < ALPSIZ ) return;
	n -= ALPSIZ;
	vector[n / BITSWD] |= 1 << (n % BITSWD);
}

static int chkbit(n)		/* check a bit in NUR bitvector */
register int n;
{
	if( n < ALPSIZ ) return( 1 );
	n -= ALPSIZ;
	return( (vector[n / BITSWD] & (1 << (n % BITSWD))) != 0 );
}

static clrvec() {		/* clear bitvector */
	register i;
	for( i = 0; i < VECSIZ; i++ )
		vector[i] = 0;
}

timer() {			/* NUR timer (global) */
	clock++;
	clock %= PERIOD;
	if( clock == 0 ) clrvec();
}

static int find() {		/* find and clear an entry */
	register int i;
	do {
		rover += BUMP;
		if( rover > DCTSIZ ) rover = rover - DCTSIZ;
		if( rover < ALPSIZ ) rover = rover + ALPSIZ;

	} while( ucount[rover-ALPSIZ] != 0 && chkbit(rover) );

	free( dict[rover] );	/* free the string */
	
	if( (i = left[rover] - DCTSIZ) >= 0 ) ucount[i]--;
	if( (i = right[rover] - DCTSIZ) >= 0 ) ucount[i]--;

	return( rover );
} /* find */

init() {			/* set up the world (global) */
	register i;
	register char *cp;

	clock = 1;
	rover = 0;

	for( i = 0; i < ALPSIZ; i++ ) {
		cp = malloc(2);
		*cp = i;
		cp[1] = '\0';
		dict[i] = cp;
		leng[i] = 1;
	}
	count = ALPSIZ;
	clrvec();
	prev = -1;
}

doadd(str1,str2)		/* put str1 // str2 into dict (global) */
int str1, str2;
{
	register char *np;
	register char *sp;
	int j,l;

	if( count < DCTSIZ ) j = count++; /* room in dict */
	else j = find();	/* no room, free an entry */

	sp = dict[str1];
	l = strlen(sp) + strlen(dict[str2]);
	np = malloc(l+1);
	dict[j] = np;
	while( (*np = *sp++) != '\0' ) np++; /* copy str1 */
	sp = dict[str2];
	while( (*np++ = *sp++) != '\0' ) ; /* append str2 */

/*	printf("'%s' '%s' %d '%s'\n",dict[str1],dict[str2],l,dict[j]); */

	j -= ALPSIZ;
	leng[j] = l;
	ucount[j] = 0;
	left[j] = str1;
	right[j] = str2;
	if( str1 >= ALPSIZ ) ucount[str1 - ALPSIZ]++;
	if( str2 >= ALPSIZ ) ucount[str2 - ALPSIZ]++;
	return( j + ALPSIZ );
} /* doadd */

dump() {			/* for debugging (global) */
	int i;
	for( i = ALPSIZ; i < count; i++ )
		printf("%d: %s\n", i, dict[i]);
} /* dump */
